/* This file is part of Lcrzo.
*/

/*-------------------------------------------------------------*/
#if defined LCRZODEF_SYSTEM_Solaris
 #define LCRZO_DEBUG_CAST_MPROTECT (lcrzo_string) 
#else
 #define LCRZO_DEBUG_CAST_MPROTECT
#endif

/*-------------------------------------------------------------*/
/* If PAGESIZE is not defined on the system (which might be ok
   because it is defined at running time), just use a hard coded
   value of 4096 for example. If this value is bad, it will be
   detected in lcrzo_debug_memprotect_alloc. */
#ifdef LCRZODEF_SYSTEM_Linux
 #include <sys/mman.h>
 #include <asm/page.h>
 #define LCRZO_PRIV_PAGESIZE PAGE_SIZE
#elif defined LCRZODEF_SYSTEM_FreeBSD
 #include <sys/mman.h>
 #include <machine/param.h>
 #define LCRZO_PRIV_PAGESIZE PAGE_SIZE
#elif defined LCRZODEF_SYSTEM_OpenBSD
 #include <sys/mman.h>
 #include <machine/param.h>
 #define LCRZO_PRIV_PAGESIZE PAGE_SIZE
#elif defined LCRZODEF_SYSTEM_Solaris
 #include <sys/mman.h>
 #include <sys/param.h>
 #define LCRZO_PRIV_PAGESIZE PAGESIZE
#elif defined LCRZODEF_SYSTEM_Unsupported
 #include <sys/mman.h>
 #define LCRZO_PRIV_PAGESIZE 4096 
#elif defined LCRZODEF_SYSTEM_Windows
 #define LCRZO_PRIV_PAGESIZE 4096
#endif

/*-------------------------------------------------------------*/
#if defined LCRZODEF_SYSTEM_Unix
int lcrzo_debug_memprotect_protect(lcrzo_ptr ptr,
				   lcrzo_uint32 ptrsize,
				   lcrzo_bool allowaccess)
{ lcrzo_ptr protectedptr;
  lcrzo_uint32 protectedpages, tmpvar;
  int prottype, ret;

  protectedpages=ptrsize/LCRZO_PRIV_PAGESIZE;

  tmpvar=LCRZO_PRIV_PAGESIZE-1;
  protectedptr=(lcrzo_ptr)((lcrzo_uintmax)((lcrzo_data)ptr+tmpvar)&(~tmpvar));

  if (allowaccess)
  { prottype=PROT_READ|PROT_WRITE|PROT_EXEC;
  }
  else
  { prottype=PROT_NONE;
  }

  ret=mprotect(LCRZO_DEBUG_CAST_MPROTECT protectedptr, 
	       protectedpages*LCRZO_PRIV_PAGESIZE, prottype);
  if (ret<0) return(LCRZO_ERR_FUMPROTECT);

  return(LCRZO_ERR_OK);
}
#elif defined LCRZODEF_SYSTEM_Windows
int lcrzo_debug_memprotect_protect(lcrzo_ptr ptr,
				   lcrzo_uint32 ptrsize,
				   lcrzo_bool allowaccess)
{ DWORD oldpro, pro;
  BOOL err;

  if (allowaccess)
  { pro=PAGE_READWRITE;
  }
  else
  { pro=PAGE_NOACCESS;
  }

  err=VirtualProtect((lcrzo_ptr)ptr, ptrsize, pro, &oldpro);
  if (!err)
  { return(LCRZO_ERR_FUVIRTUALPROTECT);
  }

  return(LCRZO_ERR_OK);
}
#endif

/*-------------------------------------------------------------*/
int lcrzo_debug_memprotect_checkpagesize(void);
#if defined LCRZODEF_SYSTEM_Unix
int lcrzo_debug_memprotect_checkpagesize(void)
{ lcrzo_uint32 pagesize;

  pagesize=getpagesize();
  if (pagesize!=(lcrzo_uint32)LCRZO_PRIV_PAGESIZE)
  { printf("This computer has a page size of %ld.\n", pagesize);
    return(LCRZO_ERR_IEINTERNALERROR);
  }
  return(LCRZO_ERR_OK);
}
#elif defined LCRZODEF_SYSTEM_Windows
int lcrzo_debug_memprotect_checkpagesize(void)
{ SYSTEM_INFO si;
  lcrzo_uint32 pagesize;

  GetSystemInfo(&si);
  pagesize=si.dwPageSize;
  if (pagesize!=LCRZO_PRIV_PAGESIZE)
  { printf("This computer has a page size of %ld.\n", pagesize);
    return(LCRZO_ERR_IEINTERNALERROR);
  }
  return(LCRZO_ERR_OK);
}
#endif

/*-------------------------------------------------------------*/
#if defined LCRZODEF_SYSTEM_Unix
int lcrzo_debug_memprotect_alloc(lcrzo_int32 allocsize,
                                 lcrzo_ptr *pptr)
{ lcrzo_ptr ptr;
  static lcrzo_bool checked=LCRZO_FALSE;

  if (!checked)
  { lcrzo_er(lcrzo_debug_memprotect_checkpagesize());
    checked=LCRZO_TRUE;
  }

  ptr=(lcrzo_ptr)malloc(allocsize);
  if (ptr==NULL) return(LCRZO_ERR_FUMALLOC);
  
  if (pptr!=NULL) *pptr=ptr;
  else free(ptr);
  return(LCRZO_ERR_OK);
}
#elif defined LCRZODEF_SYSTEM_Windows
int lcrzo_debug_memprotect_alloc(lcrzo_int32 allocsize,
                                 lcrzo_ptr *pptr)
{ void *ptrbase, *ptr;
  static lcrzo_bool checked=LCRZO_FALSE;

  if (!checked)
  { lcrzo_er(lcrzo_debug_memprotect_checkpagesize());
    checked=LCRZO_TRUE;
  }

  ptrbase=VirtualAlloc(NULL, allocsize, MEM_RESERVE, PAGE_NOACCESS);  
  if (ptrbase==NULL)
  { return(LCRZO_ERR_FUVIRTUALALLOC);
  }
  ptr=VirtualAlloc(ptrbase, allocsize, MEM_COMMIT, PAGE_READWRITE);
  if (ptr==NULL)
  { return(LCRZO_ERR_FUVIRTUALALLOC);
  }

  if (pptr!=NULL) *pptr=ptrbase;
  else { lcrzo_er(lcrzo_debug_memprotect_free(ptrbase, allocsize)); }
  
  return(LCRZO_ERR_OK);
}
#endif

/*-------------------------------------------------------------*/
#if defined LCRZODEF_SYSTEM_Unix
int lcrzo_debug_memprotect_free(lcrzo_ptr ptr,
				lcrzo_uint32 ptrsize)
{ free(ptr);
  return(LCRZO_ERR_OK);
  ptrsize=ptrsize;
}
#elif defined LCRZODEF_SYSTEM_Windows
int lcrzo_debug_memprotect_free(lcrzo_ptr ptr,
				lcrzo_uint32 ptrsize)
{ BOOL err;

  err=VirtualFree(ptr, ptrsize, MEM_DECOMMIT); 
  if (!err)
  { return(LCRZO_ERR_FUVIRTUALFREE);
  }
  err=VirtualFree(ptr, 0, MEM_RELEASE); 
  if (!err)
  { return(LCRZO_ERR_FUVIRTUALFREE);
  }

  return(LCRZO_ERR_OK);
}
#endif
